@@ -1,4 +1,4 @@
-/* $Id: gv.cpp,v 1.64 2011/01/25 16:30:51 ellson Exp $ $Revision: 1.64 $ */
+/* $Id$ $Revision$ */
 /* vim:set shiftwidth=4 ts=8: */
 
 /*************************************************************************
@@ -20,6 +20,11 @@ extern void gv_string_writer_init(GVC_t *gvc);
 extern void gv_channel_writer_init(GVC_t *gvc);
 }
 
+#define agfindattr(x,s) agattrsym(x,s)
+#define agraphattr(g,n,s) agattr(g,AGRAPH,n,s)
+#define agnodeattr(g,n,s) agattr(g,AGNODE,n,s)
+#define agedgeattr(g,n,s) agattr(g,AGEDGE,n,s)
+
 static char emptystring[] = {'\0'};
 
 static GVC_t *gvc;
@@ -33,28 +38,28 @@ Agraph_t *graph(char *name)
 {
     if (!gvc)
         gv_init();
-    return agopen(name, AGRAPH);
+    return agopen(name, Agundirected, 0);
 }
 
 Agraph_t *digraph(char *name)
 {
     if (!gvc)
         gv_init();
-    return agopen(name, AGDIGRAPH);
+    return agopen(name, Agdirected, 0);
 }
 
 Agraph_t *strictgraph(char *name)
 {
     if (!gvc)
         gv_init();
-    return agopen(name, AGRAPHSTRICT);
+    return agopen(name, Agstrictundirected, 0);
 }
 
 Agraph_t *strictdigraph(char *name)
 {
     if (!gvc)
         gv_init();
-    return agopen(name, AGDIGRAPHSTRICT);
+    return agopen(name, Agstrictdirected, 0);
 }
 
 Agraph_t *readstring(char *string)
@@ -68,7 +73,7 @@ Agraph_t *read(FILE *f)
 {
     if (!gvc)
         gv_init();
-    return agread(f);
+    return agread(f, NULL);
 }
 
 Agraph_t *read(const char *filename)
@@ -81,7 +86,7 @@ Agraph_t *read(const char *filename)
         return NULL;
     if (!gvc)
         gv_init();
-    g = agread(f);
+    g = agread(f, NULL);
     fclose(f);
     return g;
 }
@@ -91,37 +96,36 @@ Agraph_t *graph(Agraph_t *g, char *name)
 {
     if (!gvc)
         gv_init();
-    return agsubg(g, name);
+    return agsubg(g, name, 1);
 }
 
 Agnode_t *node(Agraph_t *g, char *name)
 {
-    // creating a protonode is not permitted
-    if (!gvc || (name[0] == '\001' && strcmp (name, "\001proto") == 0))
+    if (!gvc)
         return NULL;
-    return agnode(g, name);
+    return agnode(g, name, 1);
 }
 
 Agedge_t *edge(Agnode_t *t, Agnode_t *h)
 {
-    // edges from/to the protonode are not permitted
-    if (!gvc || !t || !h
-      || (agnameof(t)[0] == '\001' && strcmp (agnameof(t), "\001proto") == 0)
-      || (agnameof(h)[0] == '\001' && strcmp (agnameof(h), "\001proto") == 0))
+    if (!gvc || !t || !h)
         return NULL;
-    return agedge(t->graph, t, h);
+    // edges from/to the protonode are not permitted
+    if (AGTYPE(t) == AGRAPH || AGTYPE(h) == AGRAPH)
+	return NULL;
+    return agedge(agraphof(t), t, h, NULL, 1);
 }
 
 // induce tail if necessary
 Agedge_t *edge(char *tname, Agnode_t *h)
 {
-    return edge(node(h->graph, tname), h);
+    return edge(node(agraphof(h), tname), h);
 }
 
 // induce head if necessary
 Agedge_t *edge(Agnode_t *t, char *hname)
 {
-    return edge(t, node(t->graph, hname));
+    return edge(t, node(agraphof(t), hname));
 }
 
 // induce tail/head if necessary
@@ -138,11 +142,7 @@ static char* myagxget(void *obj, Agsym_t *a)
 
     if (!obj || !a)
         return emptystring;
-#ifndef WITH_CGRAPH
-    val = agxget(obj, a->index);
-#else
     val = agxget(obj, a);
-#endif
     if (!val)
         return emptystring;
     if (a->name[0] == 'l' && strcmp(a->name, "label") == 0 && aghtmlstr(val)) {
@@ -179,15 +179,11 @@ static void myagxset(void *obj, Agsym_t *a, char *val)
         if (val[len-1] == '>') {
             hs = strdup(val+1);
                 *(hs+len-2) = '\0';
-            val = agstrdup_html(hs);
+            val = agstrdup_html(agraphof(obj),hs);
             free(hs);
         }
     }
-#ifndef WITH_CGRAPH
-    agxset(obj, a->index, val);
-#else
     agxset(obj, a, val);
-#endif
 }
 char *setv(Agraph_t *g, Agsym_t *a, char *val)
 {
@@ -211,6 +207,10 @@ char *setv(Agraph_t *g, char *attr, char *val)
 //-------------------------------------------------
 char *getv(Agnode_t *n, Agsym_t *a)
 {
+    if (!n || !a)
+        return NULL;
+    if (AGTYPE(n) == AGRAPH) // protonode   
+	return NULL;   // FIXME ??
     return myagxget(n, a);
 }
 char *getv(Agnode_t *n, char *attr)
@@ -220,14 +220,18 @@ char *getv(Agnode_t *n, char *attr)
 
     if (!n || !attr)
         return NULL;
+    if (AGTYPE(n) == AGRAPH) // protonode   
+	return NULL;   // FIXME ??
     g = agroot(agraphof(n));
-    a = agfindattr(g->proto->n, attr);
+    a = agattr(g, AGNODE, attr, NULL);
     return myagxget(n, a);
 }
 char *setv(Agnode_t *n, Agsym_t *a, char *val)
 {
     if (!n || !a || !val)
         return NULL;
+    if (AGTYPE(n) == AGRAPH) // protonode   
+	return NULL;   // FIXME ??
     myagxset(n, a, val);
     return val;
 }
@@ -238,8 +242,14 @@ char *setv(Agnode_t *n, char *attr, char *val)
 
     if (!n || !attr || !val)
         return NULL;
+    if (AGTYPE(n) == AGRAPH) { // protonode   
+	g = (Agraph_t*)n;
+    	a = agattr(g, AGNODE, attr, val); // create default attribute in psuodo protonode
+	    // FIXME? - deal with html in "label" attributes
+	return val;
+    }
     g = agroot(agraphof(n));
-    a = agfindattr(g->proto->n, attr);
+    a = agattr(g, AGNODE, attr, NULL);
     if (!a)
         a = agnodeattr(g, attr, emptystring);
     myagxset(n, a, val);
@@ -248,6 +258,10 @@ char *setv(Agnode_t *n, char *attr, char *val)
 //-------------------------------------------------
 char *getv(Agedge_t *e, Agsym_t *a)
 {
+    if (!e || !a)
+        return NULL;
+    if (AGTYPE(e) == AGRAPH) // protoedge   
+	return NULL;   // FIXME ??
     return myagxget(e, a);
 }
 char *getv(Agedge_t *e, char *attr)
@@ -257,18 +271,18 @@ char *getv(Agedge_t *e, char *attr)
 
     if (!e || !attr)
         return NULL;
+    if (AGTYPE(e) == AGRAPH) // protoedge   
+	return NULL;   // FIXME ??
     g = agraphof(agtail(e));
-#ifndef WITH_CGRAPH
-    a = agfindattr(g->proto->e, attr);
-#else
     a = agattr(g, AGEDGE, attr, NULL);
-#endif
     return myagxget(e, a);
 }
 char *setv(Agedge_t *e, Agsym_t *a, char *val)
 {
     if (!e || !a || !val)
         return NULL;
+    if (AGTYPE(e) == AGRAPH) // protoedge   
+	return NULL;   // FIXME ??
     myagxset(e, a, val);
     return val;
 }
@@ -279,16 +293,16 @@ char *setv(Agedge_t *e, char *attr, char *val)
 
     if (!e || !attr || !val)
         return NULL;
+    if (AGTYPE(e) == AGRAPH) { // protoedge   
+	g = (Agraph_t*)e;
+    	a = agattr(g, AGEDGE, attr, val); // create default attribute in pseudo protoedge
+	    // FIXME? - deal with html in "label" attributes
+	return val;
+    }
     g = agroot(agraphof(agtail(e)));
-#ifndef WITH_CGRAPH
-    a = agfindattr(g->proto->e, attr);
-    if (!a)
-        a = agedgeattr(g, attr, emptystring);
-#else
     a = agattr(g, AGEDGE, attr, NULL);
     if (!a)
         a = agattr(g, AGEDGE, attr, emptystring);
-#endif
     myagxset(e, a, val);
     return val;
 }
@@ -297,28 +311,22 @@ Agraph_t *findsubg(Agraph_t *g, char *name)
 {
     if (!g || !name)
         return NULL;
-#ifndef WITH_CGRAPH
-    return agfindsubg(g, name);
-#else
     return agsubg(g, name, 0);
-#endif
 }
 
 Agnode_t *findnode(Agraph_t *g, char *name)
 {
     if (!g || !name)
         return NULL;
-#ifndef WITH_CGRAPH
-    return agfindnode(g, name);
-#else
     return agnode(g, name, 0);
-#endif
 }
 
 Agedge_t *findedge(Agnode_t *t, Agnode_t *h)
 {
     if (!t || !h)
         return NULL;
+    if (AGTYPE(t) == AGRAPH || AGTYPE(h) == AGRAPH)
+	return NULL;
     return agfindedge(agraphof(t), t, h);
 }
 
@@ -349,6 +357,8 @@ Agnode_t *headof(Agedge_t *e)
 {
     if (!e)
         return NULL;
+    if (AGTYPE(e) == AGRAPH)
+	return NULL;
     return aghead(e);
 }
 
@@ -356,6 +366,8 @@ Agnode_t *tailof(Agedge_t *e)
 {
     if (!e)
         return NULL;
+    if (AGTYPE(e) == AGRAPH)
+	return NULL;
     return agtail(e);
 }
 
@@ -370,6 +382,8 @@ Agraph_t *graphof(Agedge_t *e)
 {
     if (!e)
         return NULL;
+    if (AGTYPE(e) == AGRAPH)
+	return (Agraph_t*)e; /* graph of protoedge is itself recast */
     return agraphof(agtail(e));
 }
 
@@ -377,6 +391,8 @@ Agraph_t *graphof(Agnode_t *n)
 {
     if (!n)
         return NULL;
+    if (AGTYPE(n) == AGRAPH)
+	return (Agraph_t*)n;  /* graph of protonode is itself recast */
     return agraphof(n);
 }
 
@@ -392,14 +408,14 @@ Agnode_t *protonode(Agraph_t *g)
 {
     if (!g)
         return NULL;
-    return g->proto->n;
+    return (Agnode_t *)g;    // gross abuse of the type system!
 }
 
 Agedge_t *protoedge(Agraph_t *g)
 {
     if (!g)
         return NULL;
-    return g->proto->e;
+    return (Agedge_t *)g;    // gross abuse of the type system!
 }
 
 //-------------------------------------------------
@@ -413,12 +429,16 @@ char *nameof(Agnode_t *n)
 {
     if (!n)
         return NULL;
+    if (AGTYPE(n) == AGRAPH)
+	return NULL;
     return agnameof(n);
 }
 //char *nameof(Agedge_t *e)
 //{
 //    if (!e)
 //        return NULL;
+//    if (AGTYPE(e) == AGRAPH)
+//	return NULL;
 //    return agnameof(e);
 //}
 char *nameof(Agsym_t *a)
@@ -456,90 +476,27 @@ bool ok(Agsym_t *a)
 //-------------------------------------------------
 Agraph_t *firstsubg(Agraph_t *g)
 {
-    Agraph_t *mg;
-    Agnode_t *n;
-    Agedge_t *e;
-
     if (!g)
         return NULL;
-    n = g->meta_node;
-    if (!n) 
-        return NULL;
-    mg = agraphof(n);
-    if (!mg) 
-        return NULL;
-    e = agfstout(mg, n);
-    if (!e) 
-        return NULL;
-    return agusergraph(aghead(e));
+    return agfstsubg(g);
 }
 
 Agraph_t *nextsubg(Agraph_t *g, Agraph_t *sg)
 {
-    Agraph_t *mg;
-    Agnode_t *ng, *nsg;
-    Agedge_t *e;
 
     if (!g || !sg)
         return NULL;
-    ng = g->meta_node;
-    nsg = sg->meta_node;
-    if (!ng || !nsg) 
-        return NULL;
-    mg = agraphof(ng);
-    if (!mg) 
-        return NULL;
-    e = agfindedge(mg, ng, nsg);
-    if (!e) 
-        return NULL;
-    e = agnxtout(mg, e);
-    if (!e) 
-        return NULL;
-    return agusergraph(aghead(e));
+    return agnxtsubg(sg);
 }
 
 Agraph_t *firstsupg(Agraph_t *g)
 {
-    Agraph_t *mg;
-    Agnode_t *n;
-    Agedge_t *e;
-
-    if (!g)
-        return NULL;
-    n = g->meta_node;
-    if (!n) 
-        return NULL;
-    mg = agraphof(n);
-    if (!mg) 
-        return NULL;
-    e = agfstin(mg, n);
-    if (!e) 
-        return NULL;
-    return agusergraph(agtail(e));
+    return g->parent;
 }
 
 Agraph_t *nextsupg(Agraph_t *g, Agraph_t *sg)
 {
-    Agraph_t *mg;
-    Agnode_t *ng, *nsg;
-    Agedge_t *e;
-
-    if (!g || !sg)
-        return NULL;
-    ng = g->meta_node;
-    nsg = sg->meta_node;
-    if (!ng || !nsg) 
-        return NULL;
-    mg = agraphof(ng);
-    if (!mg) 
-        return NULL;
-    e = agfindedge(mg, nsg, ng);
-    if (!e) 
-        return NULL;
-    e = agnxtin(mg, e);
-    if (!e) 
-        return NULL;
-    return agusergraph(agtail(e));
+    return NULL;
 }
 
 Agedge_t *firstout(Agraph_t *g)
@@ -708,7 +665,7 @@ Agnode_t *nexttail(Agnode_t *n, Agnode_t *t)
     if (!e)
         return NULL;
     do {
-        e = agnxtout(g, e);
+        e = agnxtin(g, e);
         if (!e)
             return NULL;
     } while (agtail(e) == t);
@@ -748,9 +705,7 @@ Agsym_t *firstattr(Agraph_t *g)
     if (!g)
         return NULL;
     g = agroot(g);
-    if (dtsize(g->univ->globattr->dict) == 0)
-        return NULL;
-    return g->univ->globattr->list[0];
+    return agnxtattr(g,AGRAPH,NULL);
 }
 
 Agsym_t *nextattr(Agraph_t *g, Agsym_t *a)
@@ -760,13 +715,7 @@ Agsym_t *nextattr(Agraph_t *g, Agsym_t *a)
     if (!g || !a)
         return NULL;
     g = agroot(g);
-    for (i = 0; i < dtsize(g->univ->globattr->dict); i++)
-        if (a == g->univ->globattr->list[i])
-            break;
-    i++;
-    if (i > dtsize(g->univ->globattr->dict))
-        return NULL;
-    return g->univ->globattr->list[i];
+    return agnxtattr(g,AGRAPH,a);
 }
 
 Agsym_t *firstattr(Agnode_t *n)
@@ -776,9 +725,7 @@ Agsym_t *firstattr(Agnode_t *n)
     if (!n)
         return NULL;
     g = agraphof(n);
-    if (dtsize(g->univ->nodeattr->dict) == 0)
-        return NULL;
-    return g->univ->nodeattr->list[0];
+    return agnxtattr(g,AGNODE,NULL);
 }
 
 Agsym_t *nextattr(Agnode_t *n, Agsym_t *a)
@@ -789,13 +736,7 @@ Agsym_t *nextattr(Agnode_t *n, Agsym_t *a)
     if (!n || !a)
         return NULL;
     g = agraphof(n);
-    for (i = 0; i < dtsize(g->univ->nodeattr->dict); i++)
-        if (a == g->univ->nodeattr->list[i])
-            break;
-    i++;
-    if (i > dtsize(g->univ->nodeattr->dict))
-        return NULL;
-    return g->univ->nodeattr->list[i];
+    return agnxtattr(g,AGNODE,a);
 }
 
 Agsym_t *firstattr(Agedge_t *e)
@@ -805,9 +746,7 @@ Agsym_t *firstattr(Agedge_t *e)
     if (!e)
         return NULL;
     g = agraphof(agtail(e));
-    if (dtsize(g->univ->edgeattr->dict) == 0)
-        return NULL;
-    return g->univ->edgeattr->list[0];
+    return agnxtattr(g,AGEDGE,NULL);
 }
 
 Agsym_t *nextattr(Agedge_t *e, Agsym_t *a)
@@ -818,13 +757,7 @@ Agsym_t *nextattr(Agedge_t *e, Agsym_t *a)
     if (!e || !a)
         return NULL;
     g = agraphof(agtail(e));
-    for (i = 0; i < dtsize(g->univ->edgeattr->dict); i++)
-        if (a == g->univ->edgeattr->list[i])
-            break;
-    i++;
-    if (i > dtsize(g->univ->edgeattr->dict))
-        return NULL;
-    return g->univ->edgeattr->list[i];
+    return agnxtattr(g,AGEDGE,a);
 }
 
 bool rm(Agraph_t *g)
@@ -833,20 +766,14 @@ bool rm(Agraph_t *g)
 
     if (!g)
         return false;
-    if (g->meta_node) {
-        for (e = agfstout(g->meta_node->graph, g->meta_node); e;
-        		e = agnxtout(g->meta_node->graph, e)) {
-            rm(agusergraph(aghead(e)));
-        }
-        if (g == agroot(g)) {
-            agclose(g);
-        } else {
-            agdelete(g->meta_node->graph, g->meta_node);
-        }
-        return true;
-    }
-    fprintf(stderr, "subgraph has no meta_node\n");
-    return false;
+    Agraph_t* sg;
+    for (sg = agfstsubg (g); sg; sg = agnxtsubg (sg))
+	rm(sg);
+    if (g == agroot(g))
+	agclose(g);
+    else
+        agdelete(agroot(g), g);
+    return true;
 }
 
 bool rm(Agnode_t *n)
